# status
FAILED = "Failed"
UNNUMBERED = "Unnumbered"
SUCCESS = "Success"
LOADED = "Loaded"
NOT_LOADED = "Not Loaded"
# numbering schemes
KABAT = "kabat"
CHOTHIA = "chothia"
CHOTHIA_EXT = "chothia_ext"
MARTIN = "chothia_ext"
# servers
ABYSIS = "abysis"
