from .core import *
from .analysis.cdr_length import CDRLength
from .analysis.amino_acid_freq import AminoAcidFreq
from .analysis.sequence_alignment import SequenceAlignment
from .__about__ import __version__

VERSION = __version__
